@@ -1,10 +1,10 @@
-require 'puppet/util/package'
+require_relative '../../puppet/util/package'
 
 # Compares two version numbers.
 #
 # Prototype:
 #
-#     \$result = versioncmp(a, b)
+#     $result = versioncmp(a, b)
 #
 # Where a and b are arbitrary version strings.
 #
